@@ -47,6 +47,7 @@ OPERATIONS = {
     "kill_session": operations.KillSession,
     "poweroff_machine": operations.PoweroffMachine,
     "reboot_machine": operations.RebootMachine,
+    "rpc": operations.GenericRPC,
 }
 
 """
@@ -57,7 +58,7 @@ operations to the :class:`Manager` API.
 """
 
 
-def make_device_handler(device_params):
+def make_device_handler(device_params, ignore_errors=None):
     """
     Create a device handler object that provides device specific parameters and
     functions, which are called in various places throughout our code.
@@ -71,7 +72,7 @@ def make_device_handler(device_params):
 
     handler = device_params.get('handler', None)
     if handler:
-        return handler(device_params)
+        return handler(device_params, ignore_errors)
 
     device_name = device_params.get("name", "default")
     # Attempt to import device handler class. All device handlers are
@@ -82,7 +83,7 @@ def make_device_handler(device_params):
     dev_module_obj      = __import__(devices_module_name)
     handler_module_obj  = getattr(getattr(dev_module_obj, "devices"), device_name)
     class_obj           = getattr(handler_module_obj, class_name)
-    handler_obj         = class_obj(device_params)
+    handler_obj         = class_obj(device_params, ignore_errors)
     return handler_obj
 
 
@@ -91,6 +92,13 @@ def _extract_device_params(kwds):
 
     return device_params
 
+def _extract_errors_params(kwds):
+    errors_params = kwds.pop("errors_params", {})
+    return (
+        errors_params.get("ignore_errors", []),
+        errors_params.get("raise_mode", operations.RaiseMode.ALL)
+    )
+
 def _extract_manager_params(kwds):
     manager_params = kwds.pop("manager_params", {})
 
@@ -105,15 +113,12 @@ def _extract_nc_params(kwds):
     return nc_params
 
 def connect_ssh(*args, **kwds):
-    """
-    Initialize a :class:`Manager` over the SSH transport.
+    """Initialize a :class:`Manager` over the SSH transport.
     For documentation of arguments see :meth:`ncclient.transport.SSHSession.connect`.
 
     The underlying :class:`ncclient.transport.SSHSession` is created with
-    :data:`CAPABILITIES`. It is first instructed to
-    :meth:`~ncclient.transport.SSHSession.load_known_hosts` and then
-    all the provided arguments are passed directly to its implementation
-    of :meth:`~ncclient.transport.SSHSession.connect`.
+    :data:`CAPABILITIES`. All the provided arguments are passed directly to its
+    implementation of :meth:`~ncclient.transport.SSHSession.connect`.
 
     To customize the :class:`Manager`, add a `manager_params` dictionary in connection
     parameters (e.g. `manager_params={'timeout': 60}` for a bigger RPC timeout parameter)
@@ -124,19 +129,27 @@ def connect_ssh(*args, **kwds):
 
     A custom device handler can be provided with
     `device_params={'handler':<handler class>}` in connection parameters.
+
+    To specify errors to be be ignored in the RPC reply add
+    `errors_params={'ignore_errors': ['error pattern to ignore']}` in connection parameters.
+
+    To control the error raising mode add `raise_mode` in the `errors_params`
+    (e.g. `errors_params={'raise_mode': 0}` for ignoring all RPC errors)
+    See :class:`ncclient.operations.rpc.RaiseMode` for valid values of `raise_mode`
+
     """
     # Extract device/manager/netconf parameter dictionaries, if they were passed into this function.
     # Remove them from kwds (which should keep only session.connect() parameters).
     device_params = _extract_device_params(kwds)
     manager_params = _extract_manager_params(kwds)
     nc_params = _extract_nc_params(kwds)
+    ignore_errors, raise_mode = _extract_errors_params(kwds)
+    manager_params["raise_mode"] = raise_mode
 
-    device_handler = make_device_handler(device_params)
+    device_handler = make_device_handler(device_params, ignore_errors)
     device_handler.add_additional_ssh_connect_params(kwds)
     device_handler.add_additional_netconf_params(nc_params)
     session = transport.SSHSession(device_handler)
-    if "hostkey_verify" not in kwds or kwds["hostkey_verify"]:
-        session.load_known_hosts()
 
     try:
        session.connect(*args, **kwds)
@@ -146,17 +159,69 @@ def connect_ssh(*args, **kwds):
         raise
     return Manager(session, device_handler, **manager_params)
 
+def connect_libssh(*args, **kwargs):
+    """Initialize a :class:`Manager` over the LibSSH transport."""
+    if not hasattr(transport, 'libssh'):
+        raise ValueError("LibSSH transport is not available, install 'ssh-python' package.")
+
+    device_params = _extract_device_params(kwargs)
+    manager_params = _extract_manager_params(kwargs)
+    nc_params = _extract_nc_params(kwargs)
+    ignore_errors, raise_mode = _extract_errors_params(kwargs)
+    manager_params["raise_mode"] = raise_mode
+
+    device_handler = make_device_handler(device_params, ignore_errors)
+    device_handler.add_additional_ssh_connect_params(kwargs)
+    device_handler.add_additional_netconf_params(nc_params)
+    session = transport.libssh.LibSSHSession(device_handler)
+
+    session.connect(*args, **kwargs)
+    return Manager(session, device_handler, **manager_params)
+
+def connect_tls(*args, **kwargs):
+    """Initialize a :class:`Manager` over the TLS transport."""
+    device_params = _extract_device_params(kwargs)
+    manager_params = _extract_manager_params(kwargs)
+    nc_params = _extract_nc_params(kwargs)
+    ignore_errors, raise_mode = _extract_errors_params(kwargs)
+    manager_params["raise_mode"] = raise_mode
+
+    device_handler = make_device_handler(device_params, ignore_errors)
+    device_handler.add_additional_netconf_params(nc_params)
+    session = transport.TLSSession(device_handler)
+
+    session.connect(*args, **kwargs)
+
+    return Manager(session, device_handler, **manager_params)
+
+def connect_uds(*args, **kwargs):
+    """Initialize a :class:`Manager` over the Unix Socket transport."""
+    device_params = _extract_device_params(kwargs)
+    manager_params = _extract_manager_params(kwargs)
+    nc_params = _extract_nc_params(kwargs)
+    ignore_errors, raise_mode = _extract_errors_params(kwargs)
+    manager_params["raise_mode"] = raise_mode
+
+    device_handler = make_device_handler(device_params, ignore_errors)
+    device_handler.add_additional_netconf_params(nc_params)
+    session = transport.UnixSocketSession(device_handler)
+
+    session.connect(*args, **kwargs)
+
+    return Manager(session, device_handler, **manager_params)
 
 def connect_ioproc(*args, **kwds):
     device_params = _extract_device_params(kwds)
     manager_params = _extract_manager_params(kwds)
+    ignore_errors, raise_mode = _extract_errors_params(kwds)
+    manager_params["raise_mode"] = raise_mode
 
     if device_params:
         import_string = 'ncclient.transport.third_party.'
         import_string += device_params['name'] + '.ioproc'
         third_party_import = __import__(import_string, fromlist=['IOProc'])
 
-    device_handler = make_device_handler(device_params)
+    device_handler = make_device_handler(device_params, ignore_errors)
 
     session = third_party_import.IOProc(device_handler)
     session.connect()
@@ -171,6 +236,8 @@ def connect(*args, **kwds):
         if host == 'localhost' and device_params.get('name') == 'junos' \
                 and device_params.get('local'):
             return connect_ioproc(*args, **kwds)
+        elif kwds.pop("use_libssh", False):
+            return connect_libssh(*args, **kwds)
         else:
             return connect_ssh(*args, **kwds)
 
@@ -179,7 +246,7 @@ def call_home(*args, **kwds):
     port = kwds.get("port",4334)
     srv_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
     srv_socket.bind((host, port))
-    srv_socket.settimeout(10)
+    srv_socket.settimeout(kwds.get("timeout", 10))
     srv_socket.listen()
 
     sock, remote_host = srv_socket.accept()
@@ -187,7 +254,7 @@ def call_home(*args, **kwds):
     kwds['sock'] = sock
     return connect_ssh(*args, **kwds)
 
-class Manager(object):
+class Manager:
 
     """
     For details on the expected behavior of the operations and their
@@ -213,11 +280,11 @@ class Manager(object):
     HUGE_TREE_DEFAULT = False
     """Default for `huge_tree` support for XML parsing of RPC replies (defaults to False)"""
 
-    def __init__(self, session, device_handler, timeout=30):
+    def __init__(self, session, device_handler, timeout=30, raise_mode=operations.RaiseMode.ALL):
         self._session = session
         self._async_mode = False
         self._timeout = timeout
-        self._raise_mode = operations.RaiseMode.ALL
+        self._raise_mode = raise_mode
         self._huge_tree = self.HUGE_TREE_DEFAULT
         self._device_handler = device_handler
         self._vendor_operations = {}
@@ -289,14 +356,6 @@ class Manager(object):
                 return r
             return _missing
 
-    def rpc(self, *args, **kwds):
-        return operations.GenericRPC(self._session,
-                                     self._device_handler,
-                                     async_mode=self._async_mode,
-                                     timeout=self._timeout,
-                                     raise_mode=self._raise_mode,
-                                     huge_tree=self._huge_tree).request(*args, **kwds)
-
     def take_notification(self, block=True, timeout=None):
         """Attempt to retrieve one notification from the queue of received
         notifications.
